import os


if os.name == "nt":
    # Windows has a different strftime format for dates without leading 0
    # https://stackoverflow.com/questions/904928/python-strftime-date-without-leading-0
    DATE_FORMAT = "%#d %B %Y"
else:
    DATE_FORMAT = "%-d %B %Y"



from collections import defaultdict

from django.utils.functional import cached_property
from wagtail import hooks
from wagtail.models import Locale, Page


class PageIndex:
    """
    An in-memory index of pages to remove the need to query the database.

    Each entry in the index is a unique page by translation key, so a page
    that has been translated into different languages appears only once.
    """

    # Note: This has been designed to be as memory-efficient as possible, but it
    # hasn't been tested on a very large site yet.

    class Entry:
        """
        Represents a page in the index.
        """

        __slots__ = [
            "content_type",
            "translation_key",
            "source_locale",
            "parent_translation_key",
            "locales",
            "aliased_locales",
        ]

        def __init__(
            self,
            content_type,
            translation_key,
            source_locale,
            parent_translation_key,
            locales,
            aliased_locales,
        ):
            self.content_type = content_type
            self.translation_key = translation_key
            self.source_locale = source_locale
            self.parent_translation_key = parent_translation_key
            self.locales = locales
            self.aliased_locales = aliased_locales

        REQUIRED_PAGE_FIELDS = [
            "content_type",
            "translation_key",
            "locale",
            "path",
            "depth",
            "last_published_at",
            "latest_revision_created_at",
            "live",
        ]

        @classmethod
        def from_page_instance(cls, page):
            """
            Initialises an Entry from the given page instance.
            """
            # Get parent, but only if the parent is not the root page. We consider the
            # homepage of each langauge tree to be the roots
            if page.depth > 2:
                parent_page = page.get_parent()
            else:
                parent_page = None

            return cls(
                page.content_type,
                page.translation_key,
                page.locale,
                parent_page.translation_key if parent_page else None,
                list(
                    Page.objects.filter(
                        translation_key=page.translation_key,
                        alias_of__isnull=True,
                    ).values_list("locale", flat=True)
                ),
                list(
                    Page.objects.filter(
                        translation_key=page.translation_key,
                        alias_of__isnull=False,
                    ).values_list("locale", flat=True)
                ),
            )

    def __init__(self, pages):
        self.pages = pages

    @cached_property
    def by_translation_key(self):
        return {page.translation_key: page for page in self.pages}

    @cached_property
    def by_parent_translation_key(self):
        by_parent_translation_key = defaultdict(list)
        for page in self.pages:
            by_parent_translation_key[page.parent_translation_key].append(page)

        return dict(by_parent_translation_key.items())

    def sort_by_tree_position(self):
        """
        Returns a new index with the pages sorted in depth-first-search order
        using their parent in their respective source locale.
        """
        remaining_pages = {page.translation_key for page in self.pages}

        new_pages = []

        def _walk(translation_key):
            for page in self.by_parent_translation_key.get(translation_key, []):
                if page.translation_key not in remaining_pages:
                    continue

                remaining_pages.remove(page.translation_key)
                new_pages.append(page)
                _walk(page.translation_key)

        _walk(None)

        if remaining_pages:
            print(f"Warning: {len(remaining_pages)} orphaned pages!")

        return PageIndex(new_pages)

    def not_translated_into(self, locale):
        """
        Returns an index of pages that are not translated into the specified locale.
        This includes pages that have and don't have a placeholder
        """
        pages = [page for page in self.pages if locale.id not in page.locales]

        return PageIndex(pages)

    def __iter__(self):
        return iter(self.pages)

    @classmethod
    def from_database(cls):
        """
        Populates the index from the database.
        """
        pages = []

        for page in Page.objects.filter(alias_of__isnull=True, depth__gt=1).only(
            *PageIndex.Entry.REQUIRED_PAGE_FIELDS
        ):
            pages.append(PageIndex.Entry.from_page_instance(page))

        return PageIndex(pages)


def synchronize_tree(source_locale, target_locale, *, page_index=None):
    """
    Synchronises a locale tree with an other locale.

    This creates alias pages for pages that are not translated yet.

    Args:
        source_locale (Locale): The Locale to sync from.
        target_locale (Locale): The Locale to sync into
        page_index (PageIndex, optional): The Page index to reuse for performance. Otherwise will generate a new one.

    """
    # Build a page index
    if not page_index:
        page_index = PageIndex.from_database().sort_by_tree_position()

    # Find pages that are not translated for this locale
    # This includes locales that have a placeholder, it only excludes locales that have an actual translation
    pages_not_in_locale = page_index.not_translated_into(target_locale)

    for page in pages_not_in_locale:
        # Skip pages that do not exist in the source
        if (
            source_locale.id not in page.locales
            and source_locale.id not in page.aliased_locales
        ):
            continue

        # Fetch source from database
        model = page.content_type.model_class()
        source_page = model.objects.get(
            translation_key=page.translation_key, locale=source_locale
        )

        if target_locale.id not in page.aliased_locales:
            source_page.copy_for_translation(
                target_locale, copy_parents=True, alias=True
            )


def create_aliases_for_new_page(page):
    # Check if the source tree needs to be synchronised into any other trees
    from .models import LocaleSynchronization

    locales_to_sync_to = Locale.objects.filter(
        id__in=(
            LocaleSynchronization.objects.filter(
                sync_from_id=page.locale_id
            ).values_list("locale_id", flat=True)
        )
    )

    # Create aliases in all those locales
    for locale in locales_to_sync_to:
        new_alias = page.copy_for_translation(locale, copy_parents=True, alias=True)

        create_aliases_for_new_page(new_alias)


@hooks.register("after_create_page")
def after_create_page(request, page):
    create_aliases_for_new_page(page)



# This file contains a very lightweight implementation of RFC 72: Background workers (https://github.com/wagtail/rfcs/pull/72)
# This is only to be used by Wagtail Localize and will be replaced with the full Wagtail implementation later

from typing import Any, Callable

from django.conf import settings
from django.utils.module_loading import import_string
from typing_extensions import ParamSpec


P = ParamSpec("P")


class BaseJobBackend:
    def __init__(self, options):
        pass

    def enqueue(self, func: Callable[P, Any], args: P.args, kwargs: P.kwargs):
        raise NotImplementedError()


class ImmediateBackend(BaseJobBackend):
    def enqueue(self, func: Callable[P, Any], args: P.args, kwargs: P.kwargs):
        func(*args, **kwargs)


class DjangoRQJobBackend(BaseJobBackend):
    def __init__(self, options):
        import django_rq

        self.queue = django_rq.get_queue(options.get("QUEUE", "default"))

    def enqueue(self, func: Callable[P, Any], args: P.args, kwargs: P.kwargs):
        self.queue.enqueue(func, *args, **kwargs)


def get_backend():
    config = getattr(
        settings,
        "WAGTAILLOCALIZE_JOBS",
        {"BACKEND": "wagtail_localize.tasks.ImmediateBackend"},
    )
    backend_class = import_string(config["BACKEND"])
    return backend_class(config.get("OPTIONS", {}))


background: BaseJobBackend = get_backend()



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from wagtail.contrib.modeladmin.options import ModelAdmin
from wagtail.models import TranslatableMixin

from .helpers import TranslatableButtonHelper, TranslatablePageButtonHelper
from .views import (
    TranslatableChooseParentView,
    TranslatableCreateView,
    TranslatableDeleteView,
    TranslatableEditView,
    TranslatableHistoryView,
    TranslatableIndexView,
    TranslatableInspectView,
)


class TranslatableModelAdmin(ModelAdmin):
    index_view_class = TranslatableIndexView
    create_view_class = TranslatableCreateView
    edit_view_class = TranslatableEditView
    inspect_view_class = TranslatableInspectView
    delete_view_class = TranslatableDeleteView
    history_view_class = TranslatableHistoryView
    choose_parent_view_class = TranslatableChooseParentView

    def __init__(self, parent=None):
        super().__init__(parent)

        if "wagtail_localize.modeladmin" not in settings.INSTALLED_APPS:
            raise ImproperlyConfigured(
                'To use the TranslatableModelAdmin class "wagtail_localize.modeladmin" '
                "must be added to your INSTALLED_APPS setting."
            )

        if not issubclass(self.model, TranslatableMixin):
            raise ImproperlyConfigured(
                f"Model `{self.model}` used in translatable admin `{self.__class__}` "
                f"must subclass the `{TranslatableMixin}` class."
            )

    def get_button_helper_class(self):
        if self.button_helper_class:
            return self.button_helper_class
        if self.is_pagemodel:
            return TranslatablePageButtonHelper
        return TranslatableButtonHelper

    def get_templates(self, action="index"):
        app_label = self.opts.app_label.lower()
        model_name = self.opts.model_name.lower()
        return [
            "wagtail_localize/modeladmin/%s/%s/translatable_%s.html"
            % (app_label, model_name, action),
            "wagtail_localize/modeladmin/{}/translatable_{}.html".format(
                app_label, action
            ),
            "wagtail_localize/modeladmin/translatable_{}.html".format(action),
        ]



from bs4 import BeautifulSoup, NavigableString, Tag
from django.utils.translation import gettext_lazy as _

from wagtail_localize.strings import StringValue, lstrip_keep, rstrip_keep

from .base import BaseMachineTranslator


def language_code(code):
    return code.split("-")[0]


def translate_string(string):
    # Preserve leading/trailing whitespace. This could be a segment of a sentence.
    string, left_whitespace = lstrip_keep(string)
    string, right_whitespace = rstrip_keep(string)

    words = string.split(" ")

    return left_whitespace + " ".join(reversed(words)) + right_whitespace


def translate_html(html):
    soup = BeautifulSoup(html, "html.parser")

    def walk(soup):
        for child in soup.children:
            if isinstance(child, NavigableString):
                # Translate navigable strings
                child.string.replace_with(translate_string(child.string))

            else:
                walk(child)

    # Reverse the children
    if isinstance(soup, Tag):
        soup.contents.reverse()

    walk(soup)

    return str(soup)


class DummyTranslator(BaseMachineTranslator):
    display_name = _("Dummy translator")

    def translate(self, source_locale, target_locale, strings):
        return {string: StringValue(translate_html(string.data)) for string in strings}

    def can_translate(self, source_locale, target_locale):
        return language_code(source_locale.language_code) != language_code(
            target_locale.language_code
        )



# Generated by Django 3.1.6 on 2021-02-19 11:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("wagtail_localize", "0012_localesynchronization"),
    ]

    operations = [
        migrations.AddField(
            model_name="translationsource",
            name="schema_version",
            field=models.CharField(blank=True, max_length=255),
        ),
    ]



import django_filters

from django.contrib.contenttypes.models import ContentType
from django.utils.text import capfirst
from django.utils.translation import gettext_lazy
from django_filters.constants import EMPTY_VALUES
from django_filters.fields import ModelChoiceField
from modelcluster.fields import ParentalKey
from wagtail.admin.filters import WagtailFilterSet
from wagtail.admin.views.reports import ReportView
from wagtail.models import get_translatable_models

from wagtail_localize.models import Translation


class SourceTitleFilter(django_filters.CharFilter):
    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        return qs.filter(source__object_repr__icontains=value)


class ContentTypeModelChoiceField(ModelChoiceField):
    def label_from_instance(self, obj):
        return capfirst(obj.name)


class ContentTypeModelChoiceFilter(django_filters.ModelChoiceFilter):
    field_class = ContentTypeModelChoiceField

    def get_translatable_models(self, descendant_of=None):
        """
        Returns a list of model classes that are usable in this filter.

        When descendant_of is set, the models will be filtered to only include those that inherit from this model
        (includes the model itself if it's translatable)
        """
        translatable_models = get_translatable_models(include_subclasses=True)

        # Remove child models
        translatable_models = [
            model
            for model in translatable_models
            if not any(
                isinstance(field, ParentalKey) for field in model._meta.get_fields()
            )
        ]

        # Filter content types to only include those based on the provided type
        if descendant_of is not None:
            translatable_models = [
                model
                for model in translatable_models
                if issubclass(model, descendant_of)
            ]

        return translatable_models

    def get_queryset(self, request):
        """
        Returns QuerySet of ContentTypes to show in the filter.

        Called from django_filters.
        """
        translatable_models = self.get_translatable_models()

        # Return QuerySet of content types
        return ContentType.objects.filter(
            id__in=[
                content_type.id
                for content_type in ContentType.objects.get_for_models(
                    *translatable_models
                ).values()
            ]
        ).order_by("model")

    def filter(self, qs, value):
        """
        Filters the QuerySet to only include translations for the select content type and any descendant content types

        Called from django_filters.
        """
        if value in EMPTY_VALUES:
            return qs

        # Filter by content types that inherit from this one as well
        translatable_models = self.get_translatable_models(
            descendant_of=value.model_class()
        )
        return qs.filter(
            source__specific_content_type_id__in=[
                content_type.id
                for content_type in ContentType.objects.get_for_models(
                    *translatable_models
                ).values()
            ]
        )


class LocaleFilter(django_filters.CharFilter):
    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        return qs.filter(**{self.field_name + "__language_code": value})


class TranslationsReportFilterSet(WagtailFilterSet):
    content_type = ContentTypeModelChoiceFilter(
        field_name="source__specific_content_type", label=gettext_lazy("Content type")
    )
    source_title = SourceTitleFilter(label=gettext_lazy("Source title"))
    source_locale = LocaleFilter(field_name="source__locale")
    target_locale = LocaleFilter()

    class Meta:
        model = Translation
        fields = ["content_type", "source_title", "source_locale", "target_locale"]


class TranslationsReportView(ReportView):
    template_name = "wagtail_localize/admin/translations_report.html"
    title = gettext_lazy("Translations")
    header_icon = "site"

    filterset_class = TranslationsReportFilterSet

    def get_queryset(self):
        return Translation.objects.all()



from urllib.parse import urlencode

from django.contrib.admin.utils import quote
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured
from django.test import RequestFactory, TestCase, override_settings
from django.urls import reverse
from wagtail.models import Locale, Page
from wagtail.test.utils import WagtailTestUtils

from wagtail_localize.modeladmin import helpers
from wagtail_localize.modeladmin.options import ModelAdmin, TranslatableModelAdmin
from wagtail_localize.modeladmin.views import (
    TranslatableIndexView,
    TranslatableInspectView,
)
from wagtail_localize.models import Translation, TranslationSource
from wagtail_localize.test.models import NonTranslatableModel, TestModel, TestPage
from wagtail_localize.test.wagtail_hooks import TestModelAdmin, TestPageAdmin
from wagtail_localize.tests.utils import assert_permission_denied


def strip_user_perms():
    """
    Removes user permissions so they can still access admin and edit pages but can't submit anything for translation.
    """
    editors_group = Group.objects.get(name="Editors")
    editors_group.permissions.filter(codename="submit_translation").delete()

    for permission in Permission.objects.filter(
        content_type=ContentType.objects.get_for_model(TestModel)
    ):
        editors_group.permissions.add(permission)

    user = get_user_model().objects.get()
    user.is_superuser = False
    user.groups.add(editors_group)
    user.save()


class TestModelAdminViews(TestCase, WagtailTestUtils):
    def setUp(self):
        self.login()

        self.en_locale = Locale.objects.get()
        self.fr_locale = Locale.objects.create(language_code="fr")
        self.de_locale = Locale.objects.create(language_code="de")

        self.en_modeladmin = TestModel.objects.create(title="Test modeladmin")
        self.modeladmin_source, created = TranslationSource.get_or_create_from_instance(
            self.en_modeladmin
        )
        self.modeladmin_translation = Translation.objects.create(
            source=self.modeladmin_source, target_locale=self.fr_locale
        )
        self.modeladmin_translation.save_target(publish=True)
        self.fr_modeladmin = self.en_modeladmin.get_translation(self.fr_locale)

    def test(self):
        with self.assertRaises(ImproperlyConfigured):
            TranslatableIndexView(
                model_admin=type(
                    "NonTranslatableModelAdmin",
                    (ModelAdmin,),
                    {"model": NonTranslatableModel},
                )()
            )

    def test_index_view(self):
        response = self.client.get(
            reverse("wagtail_localize_test_testmodel_modeladmin_index")
        )
        self.assertContains(response, "Translate")

    def test_create_view(self):
        create_url = reverse("wagtail_localize_test_testmodel_modeladmin_create")
        response = self.client.get(create_url)

        # Check for correct Form Action
        self.assertContains(
            response,
            '<form action="/admin/wagtail_localize_test/testmodel/create/?locale=en"',
        )

        # Check, if other Language is available in Dropdown
        self.assertContains(
            response,
            '<a href="/admin/wagtail_localize_test/testmodel/create/?locale=de"',
        )

        # Create a new DE object if and check for locale
        response = self.client.post(create_url + "?locale=de", {"title": "New model"})
        self.assertRedirects(
            response,
            reverse("wagtail_localize_test_testmodel_modeladmin_index") + "?locale=de",
        )
        obj = TestModel.objects.last()
        self.assertEqual(obj.title, "New model")
        self.assertEqual(obj.locale, self.de_locale)

    def test_edit_view(self):
        response = self.client.get(
            reverse(
                "wagtail_localize_test_testmodel_modeladmin_edit",
                args=[self.en_modeladmin.pk],
            )
        )
        self.assertContains(
            response,
            '<a href="/admin/wagtail_localize_test/testmodel/edit/{}/?locale=fr"'.format(
                self.fr_modeladmin.pk
            ),
        )

        # Check restart translation button is displayed
        self.modeladmin_translation.enabled = False
        self.modeladmin_translation.save()
        response = self.client.get(
            reverse(
                "wagtail_localize_test_testmodel_modeladmin_edit",
                args=[self.fr_modeladmin.pk],
            )
        )
        self.assertContains(response, "Start Synced translation")

        # Check restart translation triggers correctly
        response = self.client.post(
            reverse(
                "wagtail_localize_test_testmodel_modeladmin_edit",
                args=[self.fr_modeladmin.pk],
            ),
            {"localize-restart-translation": True},
        )
        self.assertRedirects(
            response,
            reverse(
                "wagtail_localize_test_testmodel_modeladmin_edit",
                args=[self.fr_modeladmin.pk],
            ),
        )
        self.modeladmin_translation.refresh_from_db()
        self.assertTrue(self.modeladmin_translation.enabled)

        # Check existant FR translation renders the "edit_translation" view
        response = self.client.get(
            reverse(
                "wagtail_localize_test_testmodel_modeladmin_edit",
                args=[self.fr_modeladmin.pk],
            )
        )
        self.assertContains(
            response,
            "Translation of TestModel object ({}) into French".format(
                self.en_modeladmin.pk
            ),
        )

    def test_inspect_view(self):
        response = self.client.get(
            reverse(
                "wagtail_localize_test_testmodel_modeladmin_inspect",
                args=[self.en_modeladmin.pk],
            )
        )
        self.assertContains(
            response,
            '<a href="/admin/wagtail_localize_test/testmodel/inspect/{}/?locale=fr"'.format(
                self.fr_modeladmin.pk
            ),
        )
        self.assertContains(response, "Translate")
        self.assertContains(response, "Sync translated test models")

        # Create DE translation and check "Translate" button not showing
        Translation.objects.create(
            source=self.modeladmin_source, target_locale=self.de_locale
        ).save_target(publish=True)

        response = self.client.get(
            reverse(
                "wagtail_localize_test_testmodel_modeladmin_inspect",
                args=[self.fr_modeladmin.pk],
            )
        )
        self.assertNotContains(response, "Translate")
        self.assertNotContains(response, "Sync translated test models")


class TestModelAdminAdmin(TestCase, WagtailTestUtils):
    def setUp(self):
        self.login()
        self.model_admin = TestModelAdmin()
        self.page_admin = TestPageAdmin()

    def test(self):
        with self.assertRaises(ImproperlyConfigured):
            type(
                "NonTranslatableModelAdmin",
                (TranslatableModelAdmin,),
                {"model": NonTranslatableModel},
            )()

    def test_button_helper_getter(self):
        self.assertEqual(
            self.model_admin.get_button_helper_class(),
            helpers.TranslatableButtonHelper,
        )
        self.model_admin.button_helper_class = helpers.ButtonHelper
        self.assertEqual(
            self.model_admin.get_button_helper_class(),
            helpers.ButtonHelper,
        )
        self.assertEqual(
            self.page_admin.get_button_helper_class(),
            helpers.TranslatablePageButtonHelper,
        )

    def test_get_templates(self):
        def result(action):
            return [
                "wagtail_localize/modeladmin/wagtail_localize_test/testmodel/translatable_%s.html"
                % action,
                "wagtail_localize/modeladmin/wagtail_localize_test/translatable_%s.html"
                % action,
                "wagtail_localize/modeladmin/translatable_%s.html" % action,
            ]

        self.assertEqual(self.model_admin.get_templates("index"), result("index"))
        self.assertEqual(self.model_admin.get_templates("create"), result("create"))
        self.assertEqual(self.model_admin.get_templates("edit"), result("edit"))
        self.assertEqual(self.model_admin.get_templates("inspect"), result("inspect"))
        self.assertEqual(self.model_admin.get_templates("delete"), result("delete"))
        self.assertEqual(
            self.model_admin.get_templates("choose_parent"), result("choose_parent")
        )


class TestModelAdminHelpers(TestCase, WagtailTestUtils):
    def setUp(self):
        self.user = self.login()
        self.factory = RequestFactory()

        self.model_admin = TestModelAdmin()
        self.page_admin = TestPageAdmin()

        self.en_locale = Locale.objects.get()
        self.fr_locale = Locale.objects.create(language_code="fr")
        self.de_locale = Locale.objects.create(language_code="de")

        self.en_modeladmin = TestModel.objects.create(title="Test modeladmin")
        self.modeladmin_source, created = TranslationSource.get_or_create_from_instance(
            self.en_modeladmin
        )
        self.modeladmin_translation = Translation.objects.create(
            source=self.modeladmin_source, target_locale=self.fr_locale
        )
        self.modeladmin_translation.save_target(publish=True)
        self.fr_modeladmin = self.en_modeladmin.get_translation(self.fr_locale)

        root_page = Page.objects.get(id=1)
        root_page.get_children().delete()
        root_page.refresh_from_db()

        self.en_modeladmin_page = root_page.add_child(instance=TestPage(title="Test"))
        (
            self.modeladmin_page_source,
            created,
        ) = TranslationSource.get_or_create_from_instance(self.en_modeladmin_page)
        self.modeladmin_page_translation = Translation.objects.create(
            source=self.modeladmin_page_source, target_locale=self.fr_locale
        )
        self.modeladmin_page_translation.save_target(publish=True)

    def test_button_helper(self):
        request = self.factory.get(
            reverse("wagtail_localize_test_testmodel_modeladmin_index")
        )
        request.user = self.user

        view = TranslatableIndexView(self.model_admin)
        helper = self.model_admin.get_button_helper_class()(view, request)
        buttons = helper.get_buttons_for_obj(self.en_modeladmin)
        self.assertEqual(
            buttons[-1]["classname"],
            "button button-secondary button-small",
        )

        # Check inspect view sets different button classes
        view = TranslatableInspectView(self.model_admin, str(self.en_modeladmin.pk))
        helper = self.model_admin.get_button_helper_class()(view, request)
        buttons = helper.get_buttons_for_obj(self.en_modeladmin)
        self.assertEqual(buttons[-1]["classname"], "button button-secondary")

    def test_get_translation_buttons(self):
        btns = helpers.get_translation_buttons(
            self.en_modeladmin, self.user, "/next/url/", "button-class"
        )
        self.assertEqual(
            next(btns),
            {
                "url": reverse(
                    "wagtail_localize_modeladmin:submit_translation",
                    args=[
                        self.en_modeladmin._meta.app_label,
                        self.en_modeladmin._meta.model_name,
                        quote(self.en_modeladmin.pk),
                    ],
                ),
                "label": "Translate",
                "classname": "button-class",
                "title": "Translate",
            },
        )
        self.assertEqual(
            next(btns),
            {
                "url": reverse(
                    "wagtail_localize:update_translations",
                    args=[self.modeladmin_source.pk],
                )
                + "?"
                + urlencode({"next": "/next/url/"}),
                "label": "Sync translated test models",
                "classname": "button-class",
                "title": "Sync translated test models",
            },
        )

        with self.assertRaises(StopIteration):
            next(btns)

    def test_get_translation_buttons_for_page(self):
        btns = helpers.get_translation_buttons(self.en_modeladmin_page, self.user)
        self.assertEqual(
            next(btns),
            {
                "url": reverse(
                    "wagtail_localize_modeladmin:submit_translation",
                    args=[
                        self.en_modeladmin_page._meta.app_label,
                        self.en_modeladmin_page._meta.model_name,
                        quote(self.en_modeladmin_page.pk),
                    ],
                ),
                "label": "Translate",
                "classname": "",
                "title": "Translate",
            },
        )
        self.assertEqual(
            next(btns),
            {
                "url": reverse(
                    "wagtail_localize:update_translations",
                    args=[self.modeladmin_page_source.pk],
                ),
                "label": "Sync translated test pages",
                "classname": "",
                "title": "Sync translated test pages",
            },
        )

        with self.assertRaises(StopIteration):
            next(btns)

    def test_get_translation_buttons_no_locale_to_translate_to(self):
        Translation.objects.create(
            source=self.modeladmin_source, target_locale=self.de_locale
        ).save_target(publish=True)

        btns = helpers.get_translation_buttons(self.en_modeladmin, self.user)
        self.assertEqual(
            next(btns),
            {
                "url": reverse(
                    "wagtail_localize:update_translations",
                    args=[self.modeladmin_source.pk],
                ),
                "label": "Sync translated test models",
                "classname": "",
                "title": "Sync translated test models",
            },
        )

        with self.assertRaises(StopIteration):
            next(btns)

    def test_get_translation_buttons_no_user_perms(self):
        strip_user_perms()

        self.user.refresh_from_db()
        btns = helpers.get_translation_buttons(self.en_modeladmin, self.user)

        with self.assertRaises(StopIteration):
            next(btns)


@override_settings(
    LANGUAGES=[
        ("en", "English"),
        ("fr", "French"),
        ("de", "German"),
        ("es", "Spanish"),
    ],
    WAGTAIL_CONTENT_LANGUAGES=[
        ("en", "English"),
        ("fr", "French"),
        ("de", "German"),
        ("es", "Spanish"),
    ],
)
class TestTranslateModelAdminListingButton(TestCase, WagtailTestUtils):
    def setUp(self):
        self.login()

        self.en_locale = Locale.objects.get()
        self.fr_locale = Locale.objects.create(language_code="fr")
        self.de_locale = Locale.objects.create(language_code="de")

        self.en_modeladmin = TestModel.objects.create(
            title="Test modeladmin", test_textfield="Test modeladmin"
        )
        self.fr_modeladmin = self.en_modeladmin.copy_for_translation(self.fr_locale)
        self.fr_modeladmin.save()

        self.not_translatable_modeladmin = NonTranslatableModel.objects.create()

    def test(self):
        response = self.client.get(
            reverse("wagtail_localize_test_testmodel_modeladmin_index")
        )
        self.assertContains(
            response,
            (
                f'href="/admin/localize/modeladmin/submit/wagtail_localize_test/testmodel/{self.en_modeladmin.id}/" '
                f'class="button button-secondary button-small" title="Translate">Translate</a>'
            ),
        )

    def test_hides_if_modeladmin_already_translated(self):
        de_modeladmin = self.en_modeladmin.copy_for_translation(self.de_locale)
        de_modeladmin.save()

        response = self.client.get(
            reverse("wagtail_localize_test_testmodel_modeladmin_index")
        )

        self.assertNotContains(response, "Translate")

    def test_hides_if_modeladmin_isnt_translatable(self):
        de_modeladmin = self.en_modeladmin.copy_for_translation(self.de_locale)
        de_modeladmin.save()

        response = self.client.get(
            reverse("wagtail_localize_test_testmodel_modeladmin_index")
        )

        self.assertNotContains(response, "Translate")

    def test_hides_if_user_doesnt_have_permission(self):
        strip_user_perms()

        response = self.client.get(
            reverse("wagtail_localize_test_testmodel_modeladmin_index")
        )

        self.assertNotContains(response, "Translate")


@override_settings(
    LANGUAGES=[
        ("en", "English"),
        ("fr", "French"),
        ("de", "German"),
        ("es", "Spanish"),
    ],
    WAGTAIL_CONTENT_LANGUAGES=[
        ("en", "English"),
        ("fr", "French"),
        ("de", "German"),
        ("es", "Spanish"),
    ],
)
class TestSubmitModelAdminTranslation(TestCase, WagtailTestUtils):
    def setUp(self):
        self.login()

        self.en_locale = Locale.objects.get()
        self.fr_locale = Locale.objects.create(language_code="fr")
        self.de_locale = Locale.objects.create(language_code="de")

        self.en_modeladmin = TestModel.objects.create(
            title="Test modeladmin", test_textfield="Test modeladmin"
        )

        self.not_translatable_modeladmin = NonTranslatableModel.objects.create()

    def test_get_submit_modeladmin_translation(self):
        response = self.client.get(
            reverse(
                "wagtail_localize_modeladmin:submit_translation",
                args=["wagtail_localize_test", "testmodel", self.en_modeladmin.id],
            )
        )

        self.assertEqual(response.status_code, 200)

        self.assertListEqual(
            list(response.context["form"]["locales"].field.queryset),
            [self.de_locale, self.fr_locale],
        )

        # More than one locale so show "Select all"
        self.assertFalse(response.context["form"]["select_all"].field.widget.is_hidden)

        # ModelAdmin can't have children so hide include_subtree
        self.assertTrue(
            response.context["form"]["include_subtree"].field.widget.is_hidden
        )

    def test_get_submit_modeladmin_translation_when_not_modeladmin(self):
        response = self.client.get(
            reverse(
                "wagtail_localize_modeladmin:submit_translation",
                args=["wagtailcore", "page", 1],
            ),
            # Need to follow as Django will initiall redirect to /en/admin/
            follow=True,
        )

        self.assertEqual(response.status_code, 404)

    def test_get_submit_modeladmin_translation_when_invalid_model(self):
        response = self.client.get(
            reverse(
                "wagtail_localize_modeladmin:submit_translation",
                args=["wagtailcore", "foo", 1],
            ),
            # Need to follow as Django will initiall redirect to /en/admin/
            follow=True,
        )

        self.assertEqual(response.status_code, 404)

    def test_get_submit_modeladmin_translation_when_not_translatable(self):
        response = self.client.get(
            reverse(
                "wagtail_localize_modeladmin:submit_translation",
                args=[
                    "wagtail_localize_test",
                    "nontranslatablemodel",
                    self.not_translatable_modeladmin.id,
                ],
            ),
            # Need to follow as Django will initiall redirect to /en/admin/
            follow=True,
        )

        self.assertEqual(response.status_code, 404)

    def test_get_submit_modeladmin_translation_without_permissions(self):
        strip_user_perms()

        response = self.client.get(
            reverse(
                "wagtail_localize_modeladmin:submit_translation",
                args=["wagtail_localize_test", "testmodel", self.en_modeladmin.id],
            )
        )

        assert_permission_denied(self, response)

    def test_get_submit_modeladmin_translation_when_already_translated(self):
        # Locales that have been translated into shouldn't be included
        translation = self.en_modeladmin.copy_for_translation(self.de_locale)
        translation.save()

        response = self.client.get(
            reverse(
                "wagtail_localize_modeladmin:submit_translation",
                args=["wagtail_localize_test", "testmodel", self.en_modeladmin.id],
            )
        )

        self.assertEqual(response.status_code, 200)

        self.assertListEqual(
            list(response.context["form"]["locales"].field.queryset), [self.fr_locale]
        )

        # Since there is only one locale, the "Select All" checkbox should be hidden
        self.assertTrue(response.context["form"]["select_all"].field.widget.is_hidden)

    def test_post_submit_modeladmin_translation(self):
        response = self.client.post(
            reverse(
                "wagtail_localize_modeladmin:submit_translation",
                args=["wagtail_localize_test", "testmodel", self.en_modeladmin.id],
            ),
            {"locales": [self.fr_locale.id]},
        )

        translation = Translation.objects.get()
        self.assertEqual(translation.source.locale, self.en_locale)
        self.assertEqual(translation.target_locale, self.fr_locale)
        self.assertTrue(translation.created_at)

        # The translated modeladmin should've been created
        translated_modeladmin = self.en_modeladmin.get_translation(self.fr_locale)
        self.assertEqual(translated_modeladmin.test_textfield, "Test modeladmin")

        self.assertRedirects(
            response,
            reverse(
                "wagtail_localize_test_testmodel_modeladmin_edit",
                args=[translated_modeladmin.id],
            ),
        )

    def test_post_submit_modeladmin_translation_into_multiple_locales(self):
        response = self.client.post(
            reverse(
                "wagtail_localize_modeladmin:submit_translation",
                args=["wagtail_localize_test", "testmodel", self.en_modeladmin.id],
            ),
            {"locales": [self.fr_locale.id, self.de_locale.id]},
        )

        self.assertRedirects(
            response,
            reverse(
                "wagtail_localize_test_testmodel_modeladmin_edit",
                args=[self.en_modeladmin.id],
            ),
        )

        # Check French translation
        fr_translation = Translation.objects.get(target_locale=self.fr_locale)
        self.assertEqual(fr_translation.source.locale, self.en_locale)
        self.assertTrue(fr_translation.created_at)

        # Check German translation
        de_translation = Translation.objects.get(target_locale=self.de_locale)
        self.assertEqual(de_translation.source.locale, self.en_locale)
        self.assertTrue(de_translation.created_at)

    def test_post_submit_modeladmin_translation_with_missing_locale(self):
        response = self.client.post(
            reverse(
                "wagtail_localize_modeladmin:submit_translation",
                args=["wagtail_localize_test", "testmodel", self.en_modeladmin.id],
            ),
            {"locales": []},
        )

        self.assertEqual(response.status_code, 200)
        self.assertFalse(Translation.objects.exists())
        self.assertFormError(response, "form", "locales", ["This field is required."])

    def test_post_submit_modeladmin_translation_without_permissions(self):
        strip_user_perms()

        response = self.client.post(
            reverse(
                "wagtail_localize_modeladmin:submit_translation",
                args=["wagtail_localize_test", "testmodel", self.en_modeladmin.id],
            ),
            {"locales": [self.fr_locale.id]},
        )

        assert_permission_denied(self, response)


@override_settings(
    LANGUAGES=[
        ("en", "English"),
        ("fr", "French"),
        ("de", "German"),
        ("es", "Spanish"),
    ],
    WAGTAIL_CONTENT_LANGUAGES=[
        ("en", "English"),
        ("fr", "French"),
        ("de", "German"),
        ("es", "Spanish"),
    ],
)
class TestUpdateModelAdminTranslations(TestCase, WagtailTestUtils):
    def setUp(self):
        self.login()

        self.en_locale = Locale.objects.get()
        self.fr_locale = Locale.objects.create(language_code="fr")
        self.de_locale = Locale.objects.create(language_code="de")

        # Create modeladmin object and FR translation
        self.en_modeladmin = TestModel.objects.create(
            title="Test modeladmin", test_textfield="Test modeladmin"
        )
        self.modeladmin_source, created = TranslationSource.get_or_create_from_instance(
            self.en_modeladmin
        )
        self.modeladmin_translation = Translation.objects.create(
            source=self.modeladmin_source, target_locale=self.fr_locale
        )
        self.modeladmin_translation.save_target(publish=True)
        self.fr_modeladmin = self.en_modeladmin.get_translation(self.fr_locale)

    def test_get_update_modeladmin_translation(self):
        response = self.client.get(
            reverse(
                "wagtail_localize:update_translations",
                args=[self.modeladmin_source.id],
            )
        )

        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            response.context["translations"],
            [
                {
                    "title": str(self.fr_modeladmin),
                    "locale": self.fr_locale,
                    "edit_url": reverse(
                        "wagtail_localize_test_testmodel_modeladmin_edit",
                        args=[self.fr_modeladmin.id],
                    ),
                }
            ],
        )

    def test_post_update_modeladmin_translation(self):
        self.en_modeladmin.test_textfield = "Edited modeladmin"
        self.en_modeladmin.save()

        response = self.client.post(
            reverse(
                "wagtail_localize:update_translations",
                args=[self.modeladmin_source.id],
            )
        )

        self.assertRedirects(
            response,
            reverse("wagtail_localize_test_testmodel_modeladmin_index"),
        )

        # The FR version shouldn't be updated yet
        self.fr_modeladmin.refresh_from_db()
        self.assertEqual(self.fr_modeladmin.test_textfield, "Test modeladmin")

    def test_post_update_modeladmin_translation_with_publish_translations(self):
        self.en_modeladmin.test_textfield = "Edited modeladmin"
        self.en_modeladmin.save()

        response = self.client.post(
            reverse(
                "wagtail_localize:update_translations",
                args=[self.modeladmin_source.id],
            ),
            {"publish_translations": "on"},
        )

        self.assertRedirects(
            response,
            reverse("wagtail_localize_test_testmodel_modeladmin_index"),
        )

        # The FR version should be updated
        self.fr_modeladmin.refresh_from_db()
        self.assertEqual(self.fr_modeladmin.test_textfield, "Edited modeladmin")



from django.apps import AppConfig


class WagtailLocalizeModelAdminAppConfig(AppConfig):
    name = "wagtail_localize.modeladmin"
    label = "wagtail_localize_modeladmin"
    verbose_name = "Wagtail localize modeladmin"



import json

from django.core.exceptions import PermissionDenied
from django.db import transaction
from django.http import Http404
from django.shortcuts import get_object_or_404, redirect
from django.template.response import TemplateResponse
from django.utils.translation import gettext as _
from wagtail.admin import messages
from wagtail.admin.views.pages.utils import get_valid_next_url_from_request
from wagtail.models import Page, PageLogEntry, TranslatableMixin, _copy_m2m_relations
from wagtail.signals import page_published

from wagtail_localize.models import Translation, TranslationSource


def convert_to_alias(request, page_id):
    page = get_object_or_404(Page, id=page_id, alias_of_id__isnull=True)
    if not page.permissions_for_user(request.user).can_edit():
        raise PermissionDenied

    try:
        translation_source = TranslationSource.objects.get(
            object_id=page.translation_key,
            specific_content_type=page.content_type_id,
            translations__target_locale=page.locale_id,
        )

        source_page = translation_source.get_source_instance()
    except (Page.DoesNotExist, TranslationSource.DoesNotExist):
        raise Http404

    # prevent self-aliasing
    if source_page.id == page_id:
        raise Http404

    page_to_alias = page.specific

    with transaction.atomic():
        next_url = get_valid_next_url_from_request(request)

        if request.method == "POST":
            # Sync with source page
            sync_alias(
                source_page, page_to_alias, revision=source_page.get_latest_revision()
            )

            # mark the page as an alias
            page_to_alias.alias_of_id = source_page.id
            page_to_alias.save(update_fields=["alias_of_id"], clean=False)

            # note, this logs the page title from before the sync
            page_title = page_to_alias.get_admin_display_title()
            PageLogEntry.objects.log_action(
                instance=page_to_alias,
                revision=page_to_alias.get_latest_revision(),
                action="wagtail_localize.convert_to_alias",
                user=request.user,
                data={
                    "page": {
                        "id": page_to_alias.id,
                        "title": page_title,
                    },
                    "source": {
                        "id": source_page.id,
                        "title": source_page.get_admin_display_title(),
                    },
                },
            )

            # Now clean up the old translation.
            # We are deleting rather than disabling old translations as we've just gone
            # back to an alias page, which by definition doesn't hold its own content
            try:
                translation = Translation.objects.get(
                    source=translation_source, target_locale=page_to_alias.locale_id
                )
                translation.delete()
            except Translation.DoesNotExist:
                pass

            messages.success(
                request,
                _("Page '{}' has been converted into an alias.").format(page_title),
            )

            if next_url:
                return redirect(next_url)
            return redirect("wagtailadmin_pages:edit", page_to_alias.id)

    return TemplateResponse(
        request,
        "wagtail_localize/admin/confirm_convert_to_alias.html",
        {
            "page": page_to_alias,
            "source_page": source_page,
            "next": next_url,
        },
    )


def sync_alias(source_page, alias_page, revision=None, _content_json=None):
    """
    Updates the page converted to an alias to be up-to-date with the source page.
    It will also update all aliases that follow this page with the latest content from this page.
    Note: this is a modified version of Page.update_aliases() for a single alias
    """
    source_page = source_page.specific

    # Only compute this if necessary since it's quite a heavy operation
    if _content_json is None:
        # see https://github.com/wagtail/wagtail/pull/8024
        _content_json = json.loads(source_page.to_json())

    # FIXME: update when core adds better mechanism for the exclusions
    exclude_fields = [
        "id",
        "path",
        "depth",
        "numchild",
        "url_path",
        "path",
        "index_entries",
        "postgres_index_entries",
    ]

    # Copy field content
    alias_updated = alias_page.with_content_json(_content_json)

    # Mirror the publishing status of the status page
    alias_updated.live = source_page.live
    alias_updated.has_unpublished_changes = False

    # Copy child relations
    child_object_map = source_page.copy_all_child_relations(
        target=alias_updated, exclude=exclude_fields
    )

    # Process child objects
    if child_object_map:

        def process_child_object(child_object):
            if isinstance(child_object, TranslatableMixin):
                # Child object's locale must always match the page
                child_object.locale = alias_updated.locale

        for (_rel, previous_id), child_objects in child_object_map.items():
            if previous_id is None:
                for child_object in child_objects:
                    process_child_object(child_object)
            else:
                process_child_object(child_objects)

    # Copy M2M relations
    _copy_m2m_relations(source_page, alias_updated, exclude_fields=exclude_fields)

    # Don't change the aliases slug
    # Aliases can have their own slugs so they can be siblings of the original
    alias_updated.slug = alias_page.slug
    alias_updated.set_url_path(alias_updated.get_parent())

    # Technically, aliases don't have revisions, and in Page.update_aliases() fields that
    # would normally be updated by save_revision get updated. Let's mirror that.
    alias_updated.draft_title = alias_updated.title
    alias_updated.latest_revision_created_at = source_page.latest_revision_created_at

    alias_updated.save(clean=False)

    # question: should we sent the published alias?
    if alias_updated.live and not alias_page.live:
        page_published.send(
            sender=alias_updated.specific_class,
            instance=alias_updated,
            revision=revision,
            alias=True,
        )

    # Update any aliases of that alias
    alias_page.update_aliases(revision=revision, _content=_content_json)
